<?php
/**
 * Library term meta boxes handler
 *
 * @package "All In One Meta" library
 * @since   1.0.0
 * @version 1.0.0
 */

// Prevent direct script access.
if ( !defined( 'ABSPATH' ) ) {
	die( 'No direct script access allowed' );
}

if( ! class_exists( 'AIOM_User_Metabox' ) ) {
	/**
	 * Class AIOM_User_Metabox
	 */
	final class AIOM_User_Metabox {

		/**
		 * Holds meta box ID
		 * @var string|null
		 */
		private $id = NULL;

		/**
		 * Holds meta box title
		 * @var string
		 */
		private $title = '';

		/**
		 * Hold post type: post, page or any other registered post type
		 * @var array
		 */
		private $role = array();
		
		/**
		 * Holds fields configuration
		 * @var array
		 */
		private $fields = null;
		
		/**
		 * Holds callback that provide fields structure
		 * @var
		 */
		private $fields_callback;
		
		/**
		 * Holds user fields owning status
		 * @var bool
		 */
		private $has_fields = null;
		
		/**
		 * Holds current user
		 * @var null
		 */
		private $user = null;
		
		/**
		 * Get current user
		 * @return null|\WP_User
		 */
		public function get_current_user() {
			if( null === $this->user ) {
				$this->user = wp_get_current_user();
			}
			
			return $this->user;
		}

		/**
		 * AIOM_User_Metabox constructor.
		 *
		 * @param array $args
		 * @param       $fields_callback Callback that provides fields structure
		 */
		public function __construct( $args, $fields_callback ) {

			$this->validate_args( $args );
			$this->setup_args( $args );
			$this->setup_fields_callback( $fields_callback );

			$this->hooks();
		}

		/**
		 * Parse and validate metabox configuration
		 *
		 * @param array $args
		 *
		 * @throws Exception
		 */
		private function validate_args( &$args ) {
			$args = wp_parse_args( $args, array(
				'id'        => '',
				'title'     => '',
				'role'      => array()
			) );

			if ( ! $args[ 'id' ] ) {
				throw new Exception( __( '"id" is required', 'aiom' ) );
			}

			if ( ! $args[ 'title' ] ) {
				throw new Exception( __( '"title" is required', 'aiom' ) );
			}

			$args[ 'role' ] = (array)$args[ 'role' ];
			if ( empty( $args[ 'role' ] ) ) {
				throw new Exception( __( '"role" is required', 'aiom' ) );
			}
		}

		/**
		 * Setup metabox arguments
		 *
		 * @param array $args
		 */
		private function setup_args( $args ) {
			$this->id = $args[ 'id' ];
			$this->title = $args[ 'title' ];
			$this->role = $args[ 'role' ];
		}
		
		/**
		 * Setup fields callback
		 *
		 * @param string|array $callback Callback that provides fields structure
		 */
		private function setup_fields_callback( $callback ) {
			$this->fields_callback = $callback;
		}
		
		/**
		 * Check weather User has fields
		 * @param WP_User $user User instance
		 * @return bool
		 */
		private function has_fields( $user ) {
			if( null === $this->has_fields ) {
				if( in_array( '*', $this->role ) ) {
					$has_fields = true;
				} else {
					$has_fields = false;
					foreach ( $this->role as $role ) {
						if( user_can( $user, $role ) ) {
							$has_fields = true;
							break;
						}
					}
				}
				$this->has_fields = $has_fields;
			}
			
			return $this->has_fields;
		}
		
		/**
		 * Get fields
		 * @return array
		 */
		private function get_fields(){
			if( null === $this->fields ) {
				$this->fields = is_callable( $this->fields_callback ) ? call_user_func( $this->fields_callback ) : array();
			}
			
			return apply_filters( 'aiom/user/fields', $this->fields, $this->get_current_user() );
		}

		/**
		 * Setup generic actions
		 */
		private function hooks() {
			add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ) );
			add_action( 'aiom_save_user', array( $this, 'save' ), 10, 1 );
			
			add_action( 'user_new_form', array( $this, 'render_add_form' ) );
			add_action( 'show_user_profile', array( $this, 'render_edit_form' ) );
			add_action( 'edit_user_profile', array( $this, 'render_edit_form' ) );
		}

		/**
		 * Get assets version
		 * @return string
		 */
		private function get_assets_version() {
			return '1.0.0';
		}

		/**
		 * Get relative url
		 * @return string
		 */
		private function get_assets_url() {
			$dir = __DIR__;
			$theme_url = get_template_directory_uri();

			$theme_folder = substr( $theme_url, strrpos( $theme_url, '/' ) + 1 );
			$dir_url = strtr( $dir, array( '\\' => '/', DIRECTORY_SEPARATOR => '/' ) );

			$relative_url = rtrim( substr( $dir_url, strrpos( $dir_url, $theme_folder ) + strlen( $theme_folder ) ), 'classes' );

			return $theme_url . trailingslashit( untrailingslashit( $relative_url ) );

		}

		/**
		 * Enqueue Scripts and Styles
		 */
		public function admin_enqueue_scripts() {
			
			if( ! $this->has_fields( wp_get_current_user() ) ) {
				return;
			}
			$screen = get_current_screen();
			
			if ( $screen && in_array( $screen->id, array( 'user', 'user-edit', 'profile' ) ) ) {
				$relative_url = $this->get_assets_url();

				wp_enqueue_style(
					'aiom-admin-meta-style',
					$relative_url . 'assets/css/aiom-metabox-style.css',
					array(),
					$this->get_assets_version()
				);
				wp_enqueue_script(
					'aiom-admin-meta-script',
					$relative_url . 'assets/js/aiom-metabox-script.js',
					array( 'jquery' ),
					$this->get_assets_version(),
					true
				);
				
				$types = array();
				foreach( $this->get_fields() as $tab_id => $tab ) {
					$types = array_merge( $types, array_values( wp_list_pluck( $tab[ 'fields' ], 'type', '' ) ) );
				}
				
				if ( ! class_exists( 'AIOM_Renderer' ) ) {
					require_once( 'class-aiom-renderer.php' );
				}
				
				foreach( $types as $type ) {
					$handler = AIOM_Renderer::get_field_handler( $type );
					
					$handler::enqueue();
				}
			}
		}

		/**
		 * Render meta box
		 */
		public function render_add_form() {

			$meta_key = AIOM_Config::get_tax_meta_key();

			if ( ! class_exists( 'AIOM_Renderer' ) ) {
				require_once( 'class-aiom-renderer.php' );
			}

			$renderer = new AIOM_Renderer(
				$this->get_fields(),
				array(),
				$meta_key,
				'add',
				array(
					'type' => 'user',
				),
				array(
					'title' => $this->title
				)
			);
			$renderer->render();
		}
		
		

		/**
		 * Render meta box
		 * @param WP_User $object
		 */
		public function render_edit_form( $object ) {
			
			if( ! $this->has_fields( $object ) ) {
				return;
			}
			
			$meta_key = AIOM_Config::get_user_meta_key();
			$data = (array)aiom_get_user_meta( $object->ID, $meta_key );

			// in case of absentee standalone fields in single meta data - use they real meta value
			foreach( $this->get_fields() as $tab_id => $tab_data ) {
				$standalone_fields = wp_list_filter( $tab_data[ 'fields' ], array( 'standalone' => true ) );
				foreach( $standalone_fields as $name => $field_data ) {
					if( ! isset( $data[ $name ] ) ) {
						$data[ $name ] = get_user_meta( $object->ID, $name, true );
					}
				}
			}

			if ( ! class_exists( 'AIOM_Renderer' ) ) {
				require_once( 'class-aiom-renderer.php' );
			}

			$renderer = new AIOM_Renderer(
				$this->get_fields(),
				$data,
				$meta_key,
				'edit',
				$object,
				array(
					'title' => $this->title
				)
			);
			$renderer->render();
		}

		/**
		 * Add data to data saver
		 *
		 * @param int    $user_id  User ID
		 * @param array  $data     Data to save: key->value pairs
		 */
		private function add_saver_data( $user_id, $data ) {

			if ( ! class_exists( 'AIOM_Renderer' ) ) {
				require_once( 'class-aiom-renderer.php' );
			}

			$meta_values = array();
			$standalone_values = array();
			foreach ( $this->get_fields() as $tab_id => $tab ) {
				foreach ( $tab[ 'fields' ] as $field_id => $field ) {

					$handler = AIOM_Renderer::get_field_handler( $field[ 'type' ] );
					$field = call_user_func( array( $handler, 'parse_field_args' ), $field );

					$key = isset( $field[ 'meta_key' ] ) && $field[ 'meta_key' ] ? $field[ 'meta_key' ] : $field[ 'name' ];
					$default_value = isset( $field[ 'default' ] ) ? $field[ 'default' ] : null;
					$value = isset( $data[ $key ] ) ? $data[ $key ] : $default_value;
					$callback = isset( $field[ 'callback' ] ) ? $field[ 'callback' ] : false;

					if ( is_callable( $callback ) ) {
						$value = call_user_func( $callback, $value );
					}

					/**
					 * Provide possibility to overwrite meta value for standalone field
					 */
					$value = apply_filters( 'aiom/user/save_value?field=' . $key, $value, $user_id );

					$meta_values[ $key ] = $value;

					if( isset( $field[ 'standalone' ] ) && (bool)$field[ 'standalone' ] ) {
						$standalone_values[ $key ] = $value;
					}
				}
			}
			
			/**
			 * Last chance to edit fields values
			 */
			$meta_values = apply_filters( 'aiom/user/save_values', $meta_values, $user_id );

			$saver = AIOM_Data_Saver::get_instance();
			$saver->add_standalone_data( $standalone_values );
			$saver->add_data( $meta_values );

		}

		/**
		 * Save term meta data
		 *
		 * @param int $user_id  Term ID.
		 *
		 * @return bool|int
		 */
		public function save( $user_id ) {

			if( ! wp_doing_ajax() ) {

				/** Add nonce for security and authentication. */
				$nonce_name = isset( $_POST[ 'aiom_nonce' ] ) ? $_POST[ 'aiom_nonce' ] : '';
				$nonce_action = 'aiom_nonce_action';

				/** Check if nonce is valid. */
				if ( ! wp_verify_nonce( $nonce_name, $nonce_action ) ) {
					return false;
				}

			}

			/** Check if user has permissions to save data. */
			if ( ! current_user_can( 'edit_user', $user_id ) ) {
				return false;
			}

			/** make sure we are working with valid user role */
			$user = get_user_by( 'ID', $user_id );
			if( ! $this->has_fields( $user ) ) {
				return false;
			}

			/** it's save to process */
			$meta_key = AIOM_Config::get_user_meta_key();
			$meta_data = isset( $_POST[ $meta_key ] ) ? $_POST[ $meta_key ] : array();

			/** Add data to saver: it must take care */
			$this->add_saver_data( $user_id, $meta_data );
		}

	}

}